"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Context = void 0;
const ts = require("typescript");
const paths_1 = require("../utils/paths");
const index_1 = require("../models/index");
const type_parameter_1 = require("./factories/type-parameter");
const converter_1 = require("./converter");
const nodes_1 = require("./utils/nodes");
/**
 * The context describes the current state the converter is in.
 */
class Context {
    /**
     * Create a new Context instance.
     *
     * @param converter  The converter instance that has created the context.
     * @param fileNames  A list of all files that have been passed to the TypeScript compiler.
     * @param checker  The TypeChecker instance returned by the TypeScript compiler.
     */
    constructor(converter, fileNames, checker, program) {
        this.converter = converter;
        this.fileNames = fileNames;
        this.checker = checker;
        this.program = program;
        this.visitStack = [];
        const project = new index_1.ProjectReflection(converter.name);
        this.project = project;
        this.scope = project;
        if (converter.externalPattern) {
            this.externalPattern = paths_1.createMinimatch(converter.externalPattern);
        }
    }
    /** @internal */
    get logger() {
        return this.converter.application.logger;
    }
    /**
     * Return the compiler options.
     */
    getCompilerOptions() {
        return this.converter.application.options.getCompilerOptions();
    }
    /**
     * Return the type declaration of the given node.
     *
     * @param node  The TypeScript node whose type should be resolved.
     * @returns The type declaration of the given node.
     */
    getTypeAtLocation(node) {
        let nodeType;
        try {
            nodeType = this.checker.getTypeAtLocation(node);
        }
        catch (_a) {
            // ignore
        }
        if (!nodeType) {
            if (node.symbol) {
                nodeType = this.checker.getDeclaredTypeOfSymbol(node.symbol);
            }
            else if (node.parent && node.parent.symbol) {
                nodeType = this.checker.getDeclaredTypeOfSymbol(node.parent.symbol);
            }
            else if (node.parent &&
                node.parent.parent &&
                node.parent.parent.symbol) {
                nodeType = this.checker.getDeclaredTypeOfSymbol(node.parent.parent.symbol);
            }
        }
        return nodeType;
    }
    getSymbolAtLocation(node) {
        let symbol = this.checker.getSymbolAtLocation(node);
        if (!symbol && nodes_1.isNamedNode(node)) {
            symbol = this.checker.getSymbolAtLocation(node.name);
        }
        return symbol;
    }
    expectSymbolAtLocation(node) {
        const symbol = this.getSymbolAtLocation(node);
        if (!symbol) {
            throw new Error(`Expected a symbol for node with kind ${ts.SyntaxKind[node.kind]}`);
        }
        return symbol;
    }
    resolveAliasedSymbol(symbol) {
        return symbol && ts.SymbolFlags.Alias & symbol.flags
            ? this.checker.getAliasedSymbol(symbol)
            : symbol;
    }
    /**
     * Return the current logger instance.
     *
     * @returns The current logger instance.
     */
    getLogger() {
        return this.converter.application.logger;
    }
    /**
     * Register a newly generated reflection. All created reflections should be
     * passed to this method to ensure that the project helper functions work correctly.
     *
     * @param reflection  The reflection that should be registered.
     * @param symbol  The symbol the given reflection was resolved from.
     */
    registerReflection(reflection, symbol) {
        this.project.registerReflection(reflection, symbol);
    }
    /**
     * Trigger a node reflection event.
     *
     * All events are dispatched on the current converter instance.
     *
     * @param name  The name of the event that should be triggered.
     * @param reflection  The triggering reflection.
     * @param node  The triggering TypeScript node if available.
     */
    trigger(name, reflection, node) {
        this.converter.trigger(name, this, reflection, node);
    }
    /**
     * Run the given callback with the context configured for the given source file.
     *
     * @param node  The TypeScript node containing the source file declaration.
     * @param callback  The callback that should be executed.
     */
    withSourceFile(node, callback) {
        const isExternal = this.isExternalFile(node.fileName);
        if (this.isOutsideDocumentation(node.fileName, isExternal)) {
            return;
        }
        const isDeclaration = node.isDeclarationFile;
        if (isDeclaration) {
            const lib = this.converter.getDefaultLib();
            const isLib = node.fileName.substr(-lib.length) === lib;
            if (!this.converter.includeDeclarations || isLib) {
                return;
            }
        }
        this.isExternal = isExternal;
        this.isDeclaration = isDeclaration;
        this.trigger(converter_1.Converter.EVENT_FILE_BEGIN, this.project, node);
        callback();
        this.isExternal = false;
        this.isDeclaration = false;
    }
    /**
     * Run the given callback with the scope of the context set to the given reflection.
     *
     * @param scope  The reflection that should be set as the scope of the context while the callback is invoked.
     */
    withScope(scope, ...args) {
        if (!scope || !args.length) {
            return;
        }
        const callback = args.pop();
        const parameters = args.shift();
        const oldScope = this.scope;
        const oldTypeArguments = this.typeArguments;
        const oldTypeParameters = this.typeParameters;
        this.scope = scope;
        this.typeParameters = parameters
            ? this.extractTypeParameters(parameters, args.length > 0)
            : this.typeParameters;
        this.typeArguments = undefined;
        callback();
        this.scope = oldScope;
        this.typeParameters = oldTypeParameters;
        this.typeArguments = oldTypeArguments;
    }
    /**
     * Inherit the children of the given TypeScript node to the current scope.
     *
     * @param baseNode  The node whose children should be inherited.
     * @param typeArguments  The type arguments that apply while inheriting the given node.
     * @return The resulting reflection / the current scope.
     */
    inherit(baseNode, typeArguments) {
        const wasInherit = this.isInherit;
        const oldInherited = this.inherited;
        const oldInheritParent = this.inheritParent;
        const oldTypeArguments = this.typeArguments;
        this.isInherit = true;
        this.inheritParent = baseNode;
        this.inherited = [];
        const target = this.scope;
        if (!(target instanceof index_1.ContainerReflection)) {
            throw new Error("Expected container reflection");
        }
        if (baseNode.symbol) {
            const id = this.checker.getFullyQualifiedName(baseNode.symbol);
            if (this.inheritedChildren && this.inheritedChildren.includes(id)) {
                return target;
            }
            else {
                this.inheritedChildren = this.inheritedChildren || [];
                this.inheritedChildren.push(id);
            }
        }
        if (target.children) {
            this.inherited = target.children.map((c) => c.name);
        }
        else {
            this.inherited = [];
        }
        if (typeArguments) {
            this.typeArguments = this.converter.convertTypes(this, typeArguments);
        }
        else {
            this.typeArguments = undefined;
        }
        this.converter.convertNode(this, baseNode);
        this.isInherit = wasInherit;
        this.inherited = oldInherited;
        this.inheritParent = oldInheritParent;
        this.typeArguments = oldTypeArguments;
        if (!this.isInherit) {
            delete this.inheritedChildren;
        }
        return target;
    }
    /**
     * Determines if the given file is outside of the project's generated documentation.
     * This is not, and is not intended to be, foolproof. Plugins may remove reflections
     * in the file that this method returns true for, See GH#1166 for discussion on tradeoffs.
     *
     * @param fileName
     * @internal
     */
    isOutsideDocumentation(fileName, isExternal = this.isExternalFile(fileName)) {
        return isExternal && this.converter.excludeExternals;
    }
    /**
     * Checks if the given file is external.
     *
     * @param fileName
     * @internal
     */
    isExternalFile(fileName) {
        let isExternal = !this.fileNames.includes(fileName);
        if (!isExternal && this.externalPattern) {
            isExternal = this.externalPattern.some((mm) => mm.match(fileName));
        }
        return isExternal;
    }
    /**
     * Convert the given list of type parameter declarations into a type mapping.
     *
     * @param parameters  The list of type parameter declarations that should be converted.
     * @param preserve  Should the currently set type parameters of the context be preserved?
     * @returns The resulting type mapping.
     */
    extractTypeParameters(parameters, preserve) {
        const typeParameters = {};
        if (preserve) {
            Object.assign(typeParameters, this.typeParameters);
        }
        let index = 0;
        for (const declaration of parameters) {
            const name = declaration.name.text;
            if (this.typeArguments && this.typeArguments[index]) {
                typeParameters[name] = this.typeArguments[index];
            }
            else {
                const param = type_parameter_1.createTypeParameter(this, declaration);
                if (param) {
                    typeParameters[name] = param;
                }
            }
            ++index;
        }
        return typeParameters;
    }
}
exports.Context = Context;
//# sourceMappingURL=context.js.map