"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var Converter_1;
Object.defineProperty(exports, "__esModule", { value: true });
exports.Converter = void 0;
const ts = require("typescript");
const _ = require("lodash");
const context_1 = require("./context");
const components_1 = require("./components");
const component_1 = require("../utils/component");
const utils_1 = require("../utils");
const types_1 = require("./types");
/**
 * Compiles source files using TypeScript and converts compiler symbols to reflections.
 */
let Converter = Converter_1 = class Converter extends component_1.ChildableComponent {
    /**
     * Create a new Converter instance.
     *
     * @param application  The application instance this converter relies on. The application
     *   must expose the settings that should be used and serves as a global logging endpoint.
     */
    initialize() {
        this.nodeConverters = {};
    }
    addComponent(name, componentClass) {
        const component = super.addComponent(name, componentClass);
        if (component instanceof components_1.ConverterNodeComponent) {
            this.addNodeConverter(component);
        }
        return component;
    }
    addNodeConverter(converter) {
        for (const supports of converter.supports) {
            this.nodeConverters[supports] = converter;
        }
    }
    removeComponent(name) {
        const component = super.removeComponent(name);
        if (component instanceof components_1.ConverterNodeComponent) {
            this.removeNodeConverter(component);
        }
        return component;
    }
    removeNodeConverter(converter) {
        const converters = this.nodeConverters;
        const keys = _.keys(this.nodeConverters);
        for (const key of keys) {
            if (converters[key] === converter) {
                delete converters[key];
            }
        }
    }
    removeAllComponents() {
        super.removeAllComponents();
        this.nodeConverters = {};
    }
    /**
     * Compile the given source files and create a project reflection for them.
     *
     * @param fileNames  Array of the file names that should be compiled.
     */
    convert() {
        const program = ts.createProgram(this.application.options.getFileNames(), this.application.options.getCompilerOptions());
        const checker = program.getTypeChecker();
        const context = new context_1.Context(this, program.getRootFileNames(), checker, program);
        this.trigger(Converter_1.EVENT_BEGIN, context);
        const errors = this.compile(context);
        if (errors.length) {
            return errors;
        }
        const project = this.resolve(context);
        // This should only do anything if a plugin does something bad.
        project.removeDanglingReferences();
        this.trigger(Converter_1.EVENT_END, context);
        return project;
    }
    /**
     * Analyze the given node and create a suitable reflection.
     *
     * This function checks the kind of the node and delegates to the matching function implementation.
     *
     * @param context  The context object describing the current state the converter is in.
     * @param node     The compiler node that should be analyzed.
     * @return The resulting reflection or undefined.
     */
    convertNode(context, node) {
        if (context.visitStack.includes(node)) {
            return;
        }
        const oldVisitStack = context.visitStack.slice();
        context.visitStack.push(node);
        let result;
        if (node.kind in this.nodeConverters) {
            result = this.nodeConverters[node.kind].convert(context, node);
        }
        else {
            this.application.logger.warn(`Missing converter for node with kind ${ts.SyntaxKind[node.kind]}`);
        }
        context.visitStack = oldVisitStack;
        return result;
    }
    /**
     * Convert the given TypeScript type into its TypeDoc type reflection.
     *
     * @param context  The context object describing the current state the converter is in.
     * @param node  The node whose type should be reflected.
     * @param type  The type of the node if already known.
     * @returns The TypeDoc type reflection representing the given node and type.
     */
    convertType(context, node) {
        return types_1.convertType(context, node);
    }
    /**
     * Helper function to convert multiple types at once, filtering out types which fail to convert.
     *
     * @param context
     * @param nodes
     */
    convertTypes(context, nodes = [], types = []) {
        const result = [];
        _.zip(nodes, types).forEach(([node, type]) => {
            const converted = this.convertType(context, node !== null && node !== void 0 ? node : type);
            if (converted) {
                result.push(converted);
            }
        });
        return result;
    }
    /**
     * Compile the files within the given context and convert the compiler symbols to reflections.
     *
     * @param context  The context object describing the current state the converter is in.
     * @returns An array containing all errors generated by the TypeScript compiler.
     */
    compile(context) {
        const errors = ts.getPreEmitDiagnostics(context.program);
        if (errors.length) {
            return errors;
        }
        const needsSecondPass = [];
        context.inFirstPass = true;
        for (const entry of this.application.entryPoints) {
            if (entry.endsWith(".json"))
                continue;
            const sourceFile = context.program.getSourceFile(entry.replace(/\\/g, "/"));
            if (!sourceFile) {
                this.application.logger.warn(`Unable to locate entry point: ${entry}`);
                continue;
            }
            needsSecondPass.push(sourceFile);
            this.convertNode(context, sourceFile);
        }
        context.inFirstPass = false;
        for (const file of needsSecondPass) {
            this.convertNode(context, file);
        }
        return [];
    }
    /**
     * Resolve the project within the given context.
     *
     * @param context  The context object describing the current state the converter is in.
     * @returns The final project reflection.
     */
    resolve(context) {
        this.trigger(Converter_1.EVENT_RESOLVE_BEGIN, context);
        const project = context.project;
        for (const reflection of Object.values(project.reflections)) {
            this.trigger(Converter_1.EVENT_RESOLVE, context, reflection);
        }
        this.trigger(Converter_1.EVENT_RESOLVE_END, context);
        return project;
    }
    /**
     * Return the basename of the default library that should be used.
     *
     * @returns The basename of the default library.
     */
    getDefaultLib() {
        return ts.getDefaultLibFileName(this.application.options.getCompilerOptions());
    }
};
/**
 * General events
 */
/**
 * Triggered when the converter begins converting a project.
 * The listener should implement [[IConverterCallback]].
 * @event
 */
Converter.EVENT_BEGIN = "begin";
/**
 * Triggered when the converter has finished converting a project.
 * The listener should implement [[IConverterCallback]].
 * @event
 */
Converter.EVENT_END = "end";
/**
 * Factory events
 */
/**
 * Triggered when the converter begins converting a source file.
 * The listener should implement [[IConverterNodeCallback]].
 * @event
 */
Converter.EVENT_FILE_BEGIN = "fileBegin";
/**
 * Triggered when the converter has created a declaration reflection.
 * The listener should implement [[IConverterNodeCallback]].
 * @event
 */
Converter.EVENT_CREATE_DECLARATION = "createDeclaration";
/**
 * Triggered when the converter has created a signature reflection.
 * The listener should implement [[IConverterNodeCallback]].
 * @event
 */
Converter.EVENT_CREATE_SIGNATURE = "createSignature";
/**
 * Triggered when the converter has created a parameter reflection.
 * The listener should implement [[IConverterNodeCallback]].
 * @event
 */
Converter.EVENT_CREATE_PARAMETER = "createParameter";
/**
 * Triggered when the converter has created a type parameter reflection.
 * The listener should implement [[IConverterNodeCallback]].
 * @event
 */
Converter.EVENT_CREATE_TYPE_PARAMETER = "createTypeParameter";
/**
 * Triggered when the converter has found a function implementation.
 * The listener should implement [[IConverterNodeCallback]].
 * @event
 */
Converter.EVENT_FUNCTION_IMPLEMENTATION = "functionImplementation";
/**
 * Resolve events
 */
/**
 * Triggered when the converter begins resolving a project.
 * The listener should implement [[IConverterCallback]].
 * @event
 */
Converter.EVENT_RESOLVE_BEGIN = "resolveBegin";
/**
 * Triggered when the converter resolves a reflection.
 * The listener should implement [[IConverterResolveCallback]].
 * @event
 */
Converter.EVENT_RESOLVE = "resolveReflection";
/**
 * Triggered when the converter has finished resolving a project.
 * The listener should implement [[IConverterCallback]].
 * @event
 */
Converter.EVENT_RESOLVE_END = "resolveEnd";
__decorate([
    utils_1.BindOption("name")
], Converter.prototype, "name", void 0);
__decorate([
    utils_1.BindOption("externalPattern")
], Converter.prototype, "externalPattern", void 0);
__decorate([
    utils_1.BindOption("includeDeclarations")
], Converter.prototype, "includeDeclarations", void 0);
__decorate([
    utils_1.BindOption("excludeExternals")
], Converter.prototype, "excludeExternals", void 0);
__decorate([
    utils_1.BindOption("excludeNotDocumented")
], Converter.prototype, "excludeNotDocumented", void 0);
__decorate([
    utils_1.BindOption("excludePrivate")
], Converter.prototype, "excludePrivate", void 0);
__decorate([
    utils_1.BindOption("excludeProtected")
], Converter.prototype, "excludeProtected", void 0);
Converter = Converter_1 = __decorate([
    component_1.Component({
        name: "converter",
        internal: true,
        childClass: components_1.ConverterComponent,
    })
], Converter);
exports.Converter = Converter;
//# sourceMappingURL=converter.js.map