"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createDeclaration = void 0;
const ts = require("typescript");
const index_1 = require("../../models/index");
const converter_1 = require("../converter");
const comment_js_1 = require("./comment.js");
const reference_1 = require("./reference");
/**
 * List of reflection kinds that never should be static.
 */
const nonStaticKinds = [
    index_1.ReflectionKind.Class,
    index_1.ReflectionKind.Interface,
    index_1.ReflectionKind.Namespace,
];
/**
 * List of ts kinds leading to none static merge.
 */
const nonStaticMergeKinds = [
    ts.SyntaxKind.ClassDeclaration,
    ts.SyntaxKind.ClassExpression,
    ts.SyntaxKind.InterfaceDeclaration,
];
const builtInSymbolRegExp = /^__@(\w+)$/;
function hasCommentOnParentAxis(node) {
    let currentNode = node;
    while (currentNode) {
        if (comment_js_1.getRawComment(currentNode)) {
            return true;
        }
        currentNode = currentNode.parent;
    }
    return false;
}
function shouldBeIgnoredAsNotDocumented(node, kind) {
    // never ignore modules, the project, or enum members
    if (kind === index_1.ReflectionKind.Module ||
        kind === index_1.ReflectionKind.Project ||
        kind === index_1.ReflectionKind.EnumMember) {
        return false;
    }
    // do not ignore properties of the object types, that has comment themselves, for example
    //
    // /**
    //  * has docs
    //  */
    //  export SomeType = { prop1 : string }
    //
    // same applies to the inline types for function arguments:
    //
    // function someFunc(arg1 : { prop1 : string, prop2 : number }) {...}
    //
    // The `prop1` from above should be included in the docs, even that it has no documentation
    // Note, that this does not seem to apply to classes and interfaces - for those, even the class/interface
    // has docs, we still want to exclude the undocumented properties
    // Thankfully for object literals the kind of properties seems to be set to ReflectionKind.Variable
    if (kind === index_1.ReflectionKind.Variable && hasCommentOnParentAxis(node)) {
        return false;
    }
    const hasComment = Boolean(comment_js_1.getRawComment(node));
    return !hasComment;
}
/**
 * Create a declaration reflection from the given TypeScript node.
 *
 * @param context  The context object describing the current state the converter is in. The
 *   scope of the context will be the parent of the generated reflection.
 * @param node  The TypeScript node that should be converted to a reflection.
 * @param kind  The desired kind of the reflection.
 * @param name  The desired name of the reflection.
 * @returns The resulting reflection or undefined if an error is encountered.
 */
function createDeclaration(context, node, kind, name) {
    var _a;
    if (!(context.scope instanceof index_1.ContainerReflection)) {
        throw new Error("Expected container reflection.");
    }
    const container = context.scope;
    // Ensure we have a name for the reflection
    if (!name) {
        if (node.localSymbol) {
            name = node.localSymbol.name;
        }
        else if (node.symbol) {
            name = node.symbol.name;
        }
        else {
            return;
        }
        // rename built-in symbols
        const match = builtInSymbolRegExp.exec(name);
        if (match) {
            name = `[Symbol.${match[1]}]`;
        }
        else if (kind &
            (index_1.ReflectionKind.ClassMember |
                index_1.ReflectionKind.VariableOrProperty) &&
            name === "__computed") {
            // rename computed properties
            const declName = ts.getNameOfDeclaration(node);
            const symbol = declName && context.checker.getSymbolAtLocation(declName);
            if (symbol) {
                name = context.checker.symbolToString(symbol, 
                /*enclosingDeclaration*/ undefined, ts.SymbolFlags.ClassMember);
            }
            else if (declName) {
                name = declName.getText();
            }
        }
    }
    const modifiers = ts.getCombinedModifierFlags(node);
    if (context.converter.excludeNotDocumented &&
        shouldBeIgnoredAsNotDocumented(node, kind)) {
        return;
    }
    // Test whether the node is private, when inheriting ignore private members
    const isPrivate = !!(modifiers & ts.ModifierFlags.Private);
    if (context.isInherit && isPrivate) {
        return;
    }
    // Test whether the node is static, when merging a module to a class make the node static
    let isConstructorProperty = false;
    let isStatic = false;
    if (!nonStaticKinds.includes(kind)) {
        isStatic = !!(modifiers & ts.ModifierFlags.Static);
        if (container.kind === index_1.ReflectionKind.Class) {
            if (node.parent && node.parent.kind === ts.SyntaxKind.Constructor) {
                isConstructorProperty = true;
            }
            else if (!node.parent ||
                !nonStaticMergeKinds.includes(node.parent.kind)) {
                isStatic = true;
            }
        }
    }
    // Check if we already have a child of the same kind, with the same name and static flag
    let child;
    const children = (container.children = container.children || []);
    children.forEach((n) => {
        if (n.name === name &&
            n.flags.isStatic === isStatic &&
            canMergeReflectionsByKind(n.kind, kind)) {
            child = n;
        }
    });
    if (!child) {
        // Child does not exist, create a new reflection
        child = new index_1.DeclarationReflection(name, kind, container);
        child.setFlag(index_1.ReflectionFlag.Static, isStatic);
        child.setFlag(index_1.ReflectionFlag.Private, isPrivate);
        child.setFlag(index_1.ReflectionFlag.ConstructorProperty, isConstructorProperty);
        child = setupDeclaration(context, child, node);
        if (child) {
            children.push(child);
            context.registerReflection(child, (_a = context.getSymbolAtLocation(node)) !== null && _a !== void 0 ? _a : node.symbol);
        }
    }
    else {
        // Merge the existent reflection with the given node
        child = mergeDeclarations(context, child, node, kind);
    }
    // If we have a reflection, trigger the corresponding event
    if (child) {
        context.trigger(converter_1.Converter.EVENT_CREATE_DECLARATION, child, node);
    }
    return child;
}
exports.createDeclaration = createDeclaration;
/**
 * Setup a newly created declaration reflection.
 *
 * @param context  The context object describing the current state the converter is in.
 * @param reflection  The newly created blank reflection.
 * @param node  The TypeScript node whose properties should be applies to the given reflection.
 * @returns The reflection populated with the values of the given node.
 */
function setupDeclaration(context, reflection, node) {
    const modifiers = ts.getCombinedModifierFlags(node);
    reflection.setFlag(index_1.ReflectionFlag.External, context.isExternal);
    reflection.setFlag(index_1.ReflectionFlag.Protected, !!(modifiers & ts.ModifierFlags.Protected));
    reflection.setFlag(index_1.ReflectionFlag.Public, !!(modifiers & ts.ModifierFlags.Public));
    reflection.setFlag(index_1.ReflectionFlag.Optional, !!node.questionToken);
    reflection.setFlag(index_1.ReflectionFlag.Readonly, !!(modifiers & ts.ModifierFlags.Readonly));
    if (context.isInherit &&
        (node.parent === context.inheritParent ||
            reflection.flags.isConstructorProperty)) {
        if (!reflection.inheritedFrom) {
            reflection.inheritedFrom = reference_1.createReferenceType(context, node.symbol, true);
            reflection.getAllSignatures().forEach((signature) => {
                signature.inheritedFrom = reference_1.createReferenceType(context, node.symbol, true);
            });
        }
    }
    return reflection;
}
// we should not be merging type and value with the same name,
// because TypeScript has different namespaces for these two categories
function canMergeReflectionsByKind(kind1, kind2) {
    if ((kind1 & index_1.ReflectionKind.SomeType && kind2 & index_1.ReflectionKind.SomeValue) ||
        (kind2 & index_1.ReflectionKind.SomeType && kind1 & index_1.ReflectionKind.SomeValue)) {
        return false;
    }
    return true;
}
/**
 * Merge the properties of the given TypeScript node with the pre existent reflection.
 *
 * @param context  The context object describing the current state the converter is in.
 * @param reflection  The pre existent reflection.
 * @param node  The TypeScript node whose properties should be merged with the given reflection.
 * @param kind  The desired kind of the reflection.
 * @returns The reflection merged with the values of the given node or NULL if the merge is invalid.
 */
function mergeDeclarations(context, reflection, node, kind) {
    if (reflection.kind !== kind) {
        const weights = [
            index_1.ReflectionKind.Namespace,
            index_1.ReflectionKind.Enum,
            index_1.ReflectionKind.Class,
        ];
        const kindWeight = weights.indexOf(kind);
        const childKindWeight = weights.indexOf(reflection.kind);
        if (kindWeight > childKindWeight) {
            reflection.kind = kind;
        }
    }
    if (context.isInherit &&
        (context.inherited || []).includes(reflection.name) &&
        (node.parent === context.inheritParent ||
            reflection.flags.isConstructorProperty)) {
        if (!reflection.overwrites) {
            reflection.overwrites = reference_1.createReferenceType(context, node.symbol, true);
            reflection.getAllSignatures().forEach((signature) => {
                signature.overwrites = reference_1.createReferenceType(context, node.symbol, true);
            });
        }
        return;
    }
    return reflection;
}
//# sourceMappingURL=declaration.js.map